import numpy as np


class SpikesFile:
    """
    Class that contains all the addresses and timestamps of a file.

    Attributes:
            timestamps (int[]): Timestamps of the file.
            addresses (int[]): Addresses of the file.
    Note:
            Timestamps and addresses are matched, which means that timestamps[0] is the timestamp for the spike with address addresses[0].
    """

    def __init__(self, addresses=[], timestamps=[]):
        self.addresses = addresses
        self.timestamps = timestamps
        if len(timestamps) > 0:
            self.max_ts_index = np.argmax(timestamps)
            self.max_ts = timestamps[self.max_ts_index]
            self.min_ts_index = np.argmin(timestamps)
            self.min_ts = timestamps[self.min_ts_index]
        else:
            self.max_ts_index = None
            self.max_ts = None
            self.min_ts_index = None
            self.min_ts = None


class LocalizationFile:
    """
    Class that contains all the events ant timestamps from the sound source localization model of a file.

    Attributes:
            mso_neurons_ids (int[]): Neuron's IDs of the MSO population of the file.
            mso_channels (int[]): Frequency channels associated to the MSO neuron's IDs of the file.
            mso_timestamps (int[]): Timestamps of the MSO neuron's IDs of the file.
            lso_neuron_ids (int[]): Neuron's IDs of the LSO population of the file.
            lso_channels (int[]): Frequency channels associated to the LSO neuron's IDs of the file.
            lso_timestamps (int[]): Timestamps of the LSO neuron's IDs of the file.

    Note:
            Timestamps, addresses, and neurons' ID are matched, which means that mso_timestamps[0] is the timestamp for the spike with address mso_neuron_ids[0].
    """

    def __init__(self, mso_neuron_ids=[], mso_channels=[], mso_timestamps=[], lso_neuron_ids=[], lso_channels=[],
                 lso_timestamps=[]):
        self.mso_neuron_ids = mso_neuron_ids
        self.mso_channels = mso_channels
        self.mso_timestamps = mso_timestamps
        self.lso_neuron_ids = lso_neuron_ids
        self.lso_channels = lso_channels
        self.lso_timestamps = lso_timestamps


#################################################################################
##                                                                             ##
##    Copyright C 2018  Juan P. Dominguez-Morales                              ##
##                                                                             ##
##    This file is part of pyNAVIS.                                            ##
##                                                                             ##
##    pyNAVIS is free software: you can redistribute it and/or modify          ##
##    it under the terms of the GNU General Public License as published by     ##
##    the Free Software Foundation, either version 3 of the License, or        ##
##    (at your option) any later version.                                      ##
##                                                                             ##
##    pyNAVIS is distributed in the hope that it will be useful,               ##
##    but WITHOUT ANY WARRANTY; without even the implied warranty of           ##
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the              ##
##    GNU General Public License for more details.                             ##
##                                                                             ##
##    You should have received a copy of the GNU General Public License        ##
##    along with pyNAVIS.  If not, see <http://www.gnu.org/licenses/>.         ##
##                                                                             ##
#################################################################################


import math
import struct
import csv
import numpy as np
from .functions import Functions
from .objects import SpikesFile
from .objects import LocalizationFile


class Loaders:
    """
    Functionalities for loading spiking information from different formats.
    """

    @staticmethod
    def loadAEDAT(path, settings):
        """
        Loads an AEDAT (.aedat) file.

        Parameters:
                path (string): Full path of the AEDAT file to be loaded, including name and extension.
                settings (MainSettings): Configuration parameters for the file to load.

        Returns:
                SpikesFile: SpikesFile containing all the addresses and timestamps of the file.
        """
        # Read all the file
        file = open(path, "rb")
        file_data = file.read()

        # Find last header line
        end_string = "#End Of ASCII Header\r\n"
        index = file_data.find(end_string.encode("utf-8"))
        if index != -1:
            index += len(end_string)
        else:
            index = 0

        # Raw data extraction
        num_spikes = int(math.floor(len(file_data[index:]) / (settings.address_size + settings.timestamp_size)))
        spikes_array = file_data[index:index + num_spikes * (settings.address_size + settings.timestamp_size)]

        if settings.address_size != 3 and settings.timestamp_size != 3:
            address_param = ">u" + str(settings.address_size)
            timestamp_param = ">u" + str(settings.timestamp_size)
            bytes_struct = np.dtype(address_param + ", " + timestamp_param)

            spikes = np.frombuffer(spikes_array, bytes_struct)
            addresses = spikes['f0']
            timestamps = spikes['f1']

        else:
            # Separate addresses and timestamps
            spikes_struct = np.dtype([("addresses", ">u1", settings.address_size),
                                      ("timestamps", ">u1", settings.timestamp_size)])
            spikes = np.frombuffer(spikes_array, spikes_struct)

            # Fill addresses and timestamps with zeros to reach 4-bytes per element
            address_struct = np.dtype([("zeros", ">u1", (4 - settings.address_size,)),
                                       ("addresses", ">u1", (settings.address_size,))])
            timestamp_struct = np.dtype([("zeros", ">u1", (4 - settings.timestamp_size,)),
                                         ("timestamps", ">u1", (settings.timestamp_size,))])
            filled_addresses = np.zeros(len(spikes), dtype=address_struct)
            filled_timestamps = np.zeros(len(spikes), dtype=timestamp_struct)
            filled_addresses['addresses'] = spikes['addresses']
            filled_timestamps['timestamps'] = spikes['timestamps']

            # View these filled addresses and timestamps as 4-byte ints
            addresses = filled_addresses.view(">u4")
            timestamps = filled_timestamps.view(">u4")

        spikes_file = SpikesFile(addresses, timestamps)

        # Close the file
        file.close()

        # Check correct address values and increasing timestamp order in the loaded aedat file
        _, order_is_ok, all_in_range = Functions.check_SpikesFile(spikes_file, settings)
        if not all_in_range:
            raise ValueError("Addresses are not in range. Could be due to bad decoding")
        if not order_is_ok:
            Functions.order_SpikesFile(spikes_file, settings)

        return spikes_file

    # TODO: Check the rest of loading functions

    @staticmethod
    def loadAEDATLocalization(path, settings, localization_settings):
        """
        Loads an AEDAT (.aedat) file which contains events from both the NAS model and the SOC model (sound source localization).
        
        Parameters:
                path (string): Full path of the AEDAT file to be loaded, including name and extension.
                settings (MainSettings): Configuration parameters for the file to load.
                localization_settings (LocalizationSettings): Configuration parameters of the localization module for the file to load.
        Returns:
                SpikesFile: SpikesFile containing all the addresses and timestamps of the file.
                LocalizationFile: LocalizationFile containing all the events from both the MSO and LSO models of the file.
        Raises:
                SettingsError: If settings.address_size is different than 2 and 4.

        """
        unpack_param = ">H"

        if settings.address_size == 2:
            unpack_param = ">H"
        elif settings.address_size == 4:
            unpack_param = ">L"
        else:
            print("[Loaders.loadAEDATLocalization] > SettingsError: Only address sizes implemented are 2 and 4 bytes")

        # Check the localization_settings values
        localization_settings_error = False

        # MSO start frequency channel range
        if ((localization_settings.mso_start_channel < 0) or (localization_settings.mso_start_channel >= settings.num_channels)):
            print("[Loaders.loadAEDATLocalization] > LocalizationSettingsError: MSO start frequency channel range should be in the range [0, num_channels-1]")
            localization_settings_error = True

        # MSO start frequency channel and end frequency channel
        if (localization_settings.mso_end_channel < localization_settings.mso_start_channel):
            print("[Loaders.loadAEDATLocalization] > LocalizationSettingsError: MSO start frequency channel should be lower than MSO end frequency channel")
            localization_settings_error = True

        # MSO start frequency channel and end frequency channel
        if ((localization_settings.mso_num_neurons_channel < 1) or (localization_settings.mso_num_neurons_channel > 32)):
            print("[Loaders.loadAEDATLocalization] > LocalizationSettingsError: MSO number of neurons value should be in the range [1, 32]")
            localization_settings_error = True

        # LSO start frequency channel range
        if ((localization_settings.lso_start_channel < 0) or (localization_settings.lso_start_channel >= settings.num_channels)):
            print("[Loaders.loadAEDATLocalization] > LocalizationSettingsError: LSO start frequency channel range should be in the range [0, num_channels-1]")
            localization_settings_error = True

        # LSO start frequency channel and end frequency channel
        if (localization_settings.lso_end_channel < localization_settings.lso_start_channel):
            print("[Loaders.loadAEDATLocalization] > LocalizationSettingsError: LSO start frequency channel should be lower than LSO end frequency channel")
            localization_settings_error = True

        # LSO start frequency channel and end frequency channel
        if ((localization_settings.lso_num_neurons_channel < 1) or (localization_settings.lso_num_neurons_channel > 32)):
            print("[Loaders.loadAEDATLocalization] > LocalizationSettingsError: lSO number of neurons value should be in the range [1, 32]")
            localization_settings_error = True

        if(localization_settings_error):
            return None

        with open(path, 'rb') as f:
            ## Check header ##
            p = 0
            lt = f.readline()
            while lt and lt[0] == ord("#"):
                p += len(lt)
                lt = f.readline()
            f.seek(p)

            f.seek(0, 2)
            eof = f.tell()

            num_events = math.floor((eof-p)/(settings.address_size + 4))

            f.seek(p)

            events_nas = []
            timestamps_nas = []

            neuron_ids_mso = []
            channels_mso =  []
            timestamps_mso = []

            neuron_ids_lso = []
            channels_lso =  []
            timestamps_lso = []

            ## Read file ##
            i = 0
            total_number_events_counter = 0
            invalid_localization_data_counter = 0
            try:
                while 1:
                    # Read a word and unpack the event data
                    buff = f.read(settings.address_size)
                    ev = struct.unpack(unpack_param, buff)[0]
                    total_number_events_counter = total_number_events_counter + 1
                    # Read a word and unpack the event timestamp
                    buff = f.read(4)
                    ts = struct.unpack('>L', buff)[0]

                    # Check if the event is a NAS event of SOC event
                    auditory_model = (ev & 0x8000) >> 15

                    if auditory_model == 0:
                        # NAS event
                        events_nas.append(ev)
                        timestamps_nas.append(ts)
                    elif auditory_model == 1:
                        # Localization event

                        # Set the valid data flag to true
                        valid_localization_data = True

                        # Apply a mask to obtain the correct values and check them
                        neuron_id = (ev & 0x3E00) >> 9
                        if neuron_id < 0 or neuron_id >= localization_settings.mso_num_neurons_channel:
                            valid_localization_data = False
                        freq_channel = (ev & 0x00FE) >> 1
                        if freq_channel < localization_settings.mso_start_channel or freq_channel > localization_settings.mso_end_channel:
                            valid_localization_data = False

                        xso_type = (ev & 0x4000) >> 14

                        if xso_type == 0:
                            # MSO event
                            if valid_localization_data:
                                neuron_ids_mso.append(neuron_id)
                                channels_mso.append(freq_channel)
                                timestamps_mso.append(ts)
                            else:
                                invalid_localization_data_counter = invalid_localization_data_counter + 1
                        elif xso_type == 1:
                            # LSO event
                            neuron_ids_lso.append(neuron_id)
                            channels_lso.append(freq_channel)
                            timestamps_lso.append(ts)
                        else:
                            # Other case
                            print("[Loaders.loadAEDATLocalization] > DataError: MSO/LSO type not recognized!")
                    else:
                        # Other case
                        print("[Loaders.loadAEDATLocalization] > DataError: Auditory model not recognized!")

                    i += 1
            except Exception as inst:
                pass
        spikes_file = SpikesFile([], [])
        spikes_file.addresses = events_nas
        spikes_file.timestamps = timestamps_nas

        localization_file = LocalizationFile([], [], [], [], [], [])
        localization_file.mso_neuron_ids = neuron_ids_mso
        localization_file.mso_channels = channels_mso
        localization_file.mso_timestamps = timestamps_mso
        localization_file.lso_neuron_ids = neuron_ids_lso
        localization_file.lso_channels = channels_lso
        localization_file.lso_timestamps = timestamps_lso
        # Let the user know if there were dumped events
        if invalid_localization_data_counter > 0:
            print("[Loaders.loadAEDATLocalization] > DataWarning: " + str(invalid_localization_data_counter) + " of " + str(total_number_events_counter) + " were dumped due to invalid unpacked data!")
        return spikes_file, localization_file

    @staticmethod
    def loadCSV(path, delimiter=','):
        """
        Loads a Comma-Separated Values (.csv) file.
        
        Parameters:
                path (string): Full path of the CSV file to be loaded, including name and extension.
                delimiter (char): Delimiter to use in the CSV file.

        Returns:
                SpikesFile: SpikesFile containing all the addresses and timestamps of the file.

        Note:
                The CSV file should contain one line per event, and the information in each line should be: address, timestamp

        """
        addresses = []
        timestamps = []


        with open(path) as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=delimiter)
            for row in csv_reader:
                addresses.append(int(row[0]))
                timestamps.append(int(row[1]))

        spikes_file = SpikesFile([], [])
        spikes_file.addresses = addresses
        spikes_file.timestamps = timestamps
        return spikes_file

    @staticmethod
    def loadCSVLocalization(path, delimiter=','):
        """
        Loads a Comma-Separated Values (.csv) file which contains events from both the NAS model and the SOC model (sound source localization).
        
        Parameters:
                path (string): Full path of the CSV file to be loaded, including name and extension.
                delimiter (char): Delimiter to use in the CSV file.

        Returns:
                SpikesFile: SpikesFile containing all the addresses and timestamps of the file.
                LocalizationFile: LocalizationFile containing all the events from both the MSO and LSO models of the file.

        Note:
                The CSV file should contain one line per event, and the information in each line should be: address, timestamp

                The CSV format should be: address, timestamp, auditory_model, xso_type, neuron_id.

        """
        addresses_nas = []
        timestamps_nas = []

        neuron_ids_mso = []
        channels_mso =  []
        timestamps_mso = []

        neuron_ids_lso = []
        channels_lso =  []
        timestamps_lso = []


        with open(path) as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=delimiter)
            for row in csv_reader:

                try:
                    auditory_model = int(row[2])
                except Exception:
                    auditory_model = 0
                    pass

                address = int(row[0])
                timestamp = row[1]

                # Check if the timestamp contains any time reference (from simulation)
                timeref = "ps"
                if timeref in timestamp:
                    # Remove string "ps" and convert the timestamps from picoseconds to microseconds
                    timestamp = timestamp.replace(" ps", "")
                    timestamp = int(timestamp)
                    timestamp = timestamp * 1.0e-6

                if auditory_model == 0:
                    # NAS event
                    addresses_nas.append(address)
                    timestamps_nas.append(int(timestamp))
                elif auditory_model == 1:
                    # Localization event
                    xso_type = int(row[3])
                    neuron_id = int(row[4])

                    freq_channel = address #>> 1
                    if xso_type == 0:
                        # MSO event
                        neuron_ids_mso.append(neuron_id)
                        channels_mso.append(freq_channel)
                        timestamps_mso.append(timestamp)
                    elif xso_type == 1:
                        # LSO event
                        neuron_ids_lso.append(neuron_id)
                        channels_lso.append(freq_channel)
                        timestamps_lso.append(timestamp)
                    else:
                        # Other case
                        print("[Loaders.loadCSVLocalization] > DataError: MSO/LSO type not recognized!")
                else:
                    # Other case
                    print("[Loaders.loadCSVLocalization] > DataError: Auditory model not recognized!")

        spikes_file = SpikesFile([], [])
        spikes_file.addresses = addresses_nas
        spikes_file.timestamps = timestamps_nas

        localization_file = LocalizationFile([], [], [], [], [], [])
        localization_file.mso_neuron_ids = neuron_ids_mso
        localization_file.mso_channels = channels_mso
        localization_file.mso_timestamps = timestamps_mso
        localization_file.lso_neuron_ids = neuron_ids_lso
        localization_file.lso_channels = channels_lso
        localization_file.lso_timestamps = timestamps_lso

        return spikes_file, localization_file

    @staticmethod
    def loadZynqGrabberData(path, settings, localization_settings):
        """
        Loads a text (.txt) file with EAR events collected by the zynqGrabber.
        
        Parameters:
                path (string): Full path of the CSV file to be loaded, including name and extension.
                settings (MainSettings): Configuration parameters for the file to load.
                localization_settings (LocalizationSettings): Configuration parameters of the localization module for the file to load.

        Returns:
                spikes_file: SpikesFile containing all the addresses and timestamps of the file.
                localization_file: LocalizationFile containing all the events from both the MSO and LSO models of the file.
        """

        addresses = []
        timestamps = []

        neuron_ids_mso = []
        channels_mso =  []
        timestamps_mso = []

        neuron_ids_lso = []
        channels_lso =  []
        timestamps_lso = []

        txt_file = open(path, 'r')
        txt_lines = txt_file.readlines()

        count = 0
        for line in txt_lines:
            event = line.strip().split(',')
            decoded_events_timestamps      = float(event[0])
            decoded_events_auditory_models = int(event[1])           # 0 if the event comes from the NAS, 1 for the SOC model
            decoded_events_channels        = int(event[2])           # 0 left, 1 right
            decoded_events_xso_types       = int(event[3])           # 0 for MSO, 1 for LSO
            decoded_events_neuron_ids      = int(event[4])           # Between 0 and 15
            decoded_events_freq_ch_addrs   = int(event[5])           # Between 0 and 32
            decoded_events_polarities      = int(event[6])           # 0 pos, 1 neg

            # It could be either NAS (auditory_models = 0) or SOC events (auditory_models = 1)
            if decoded_events_auditory_models == 0:
                # NAS event
                timestamps.append(int(decoded_events_timestamps))
                addresses.append(int(decoded_events_freq_ch_addrs*(1+settings.on_off_both) + decoded_events_polarities +  settings.num_channels*decoded_events_channels*(1+settings.on_off_both)))

            elif decoded_events_auditory_models == 1:
                # It could be either MSO (xso_type = 0) or LSO events (xso_type = 1)
                if decoded_events_xso_types == 0:
                    # MSO event
                    timestamps_mso.append(int(decoded_events_timestamps))
                    channels_mso.append(int(decoded_events_freq_ch_addrs))
                    neuron_ids_mso.append(int(decoded_events_neuron_ids))

                elif decoded_events_xso_types == 1:
                    # LSO event
                    timestamps_lso.append(int(decoded_events_timestamps))
                    channels_lso.append(int(decoded_events_freq_ch_addrs))
                    neuron_ids_lso.append(int(decoded_events_neuron_ids))
                else:
                    # Other case
                    print("[Loaders.loadZynqGrabberData] > DataError: MSO/LSO type not recognized!")

            else:
                # Other case
                print("[Loaders.loadZynqGrabberData] > DataError: Auditory model not recognized!")

        spikes_file = SpikesFile([], [])
        spikes_file.addresses = addresses
        spikes_file.timestamps = timestamps

        localization_file = LocalizationFile([], [], [], [], [], [])
        localization_file.mso_neuron_ids = neuron_ids_mso
        localization_file.mso_channels = channels_mso
        localization_file.mso_timestamps = timestamps_mso
        localization_file.lso_neuron_ids = neuron_ids_lso
        localization_file.lso_channels = channels_lso
        localization_file.lso_timestamps = timestamps_lso

        return spikes_file, localization_file


import datetime
import os

import matplotlib
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages

from .functions import Functions
from .loaders import Loaders
from .objects import LocalizationFile
from .objects import SpikesFile
from .plots import Plots


class ReportFunctions:

    @staticmethod
    def PDF_report(spikes_file, settings, output_path,
                   plots=["Spikegram", "Sonogram", "Histogram", "Average activity", "Difference between L/R"],
                   add_localization_report=False, localization_file=None, localization_settings=None,
                   localization_plots=["MSO spikegram", "MSO heatmap", "MSO histogram", "MSO localization"],
                   vector=False, verbose=False):
        """
		Generates a PDF report with the spikegram, sonogram, histogram, average activity and difference between L/R plots obtained from the input SpikesFile or path containing SpikeFiles.

		Parameters:
				spikes_file (SpikesFile or string): File or path to use.
				settings (MainSettings): Configuration parameters for the input file.
				output_path (string): Destination path.
				plots (string[]): List to select the plots to be included in the PDF report.
				add_localization_report (boolean, optional): If True, the localization plots will be included in the PDF report.
				localization_file (LocalizationFile, optional): If add_localization_report is set to True, this parameter is mandatory, and it should contain the localization information.
				localization_settings (LocalizationSettings, optional): If add_localization_report is set to True, this parameter is mandatory, and it should contain the localization settings.
				localization_plots (string[], optional): If add_localization_report is set to True, this parameter is mandatory, and it should contain the list of localization plots.
				vector (boolean, optional): Set to True if you want the Spikegram plot vectorized. Note: this may make your PDF heavy.
				verbose (boolean, optional): Set to True if you want the execution time of the function to be printed.

		Returns:
				None.

		Notes:
				If the path used as input is a folder instead of a spikes file, the PDF report is generated for every spikes file contained in the folder.
		"""

        # Set a non-interactive backend for matplotlib
        matplotlib.use('pdf')

        if isinstance(spikes_file, str):

            spikes_file_extension = os.path.splitext(spikes_file)

            if spikes_file_extension[1] == ".aedat":
                if add_localization_report == False:
                    spikes_file = Loaders.loadAEDAT(spikes_file, settings)
                elif add_localization_report != False and localization_file != None and localization_settings != None:
                    spikes_file, localization_file = Loaders.loadAEDATLocalization(spikes_file, settings,
                                                                                   localization_settings)
                else:
                    print("[Functions.PDF_report] > ParametersError: the input parameters are not correct.")
                    return None
            elif spikes_file_extension[1] == ".csv":
                if add_localization_report == False:
                    spikes_file = Loaders.loadCSV(spikes_file, delimiter=',')
                elif add_localization_report != False and localization_file != None and localization_settings != None:
                    spikes_file, localization_file = Loaders.loadCSVLocalization(spikes_file, delimiter=',')
                else:
                    print("[Functions.PDF_report] > ParametersError: the input parameters are not correct.")
                    return None
            elif spikes_file_extension[1] == ".txt":
                spikes_file, localization_file = Loaders.loadZynqGrabberData(spikes_file, settings,
                                                                             localization_settings)
            else:
                print("[Functions.PDF_report] > InputFileExtensionError: the extension of the input file is not valid.")
                return None
            if spikes_file.min_ts != 0:
                Functions.adapt_timestamps(spikes_file, settings)
            if add_localization_report == True:
                localization_file.timestamps = Functions.adapt_timestamps(localization_file.timestamps, settings)

        if isinstance(spikes_file, SpikesFile):
            pdf = PdfPages(output_path)
            figures = list()

            # Spikegram
            if "Spikegram" in plots:
                spk_fig = Plots.spikegram(spikes_file, settings)
                figures.append(spk_fig)

            # Sonogram
            if "Sonogram" in plots:
                sng_fig = Plots.sonogram(spikes_file, settings)
                figures.append(sng_fig)

            # Histogram
            if "Histogram" in plots:
                _, hst_fig = Plots.histogram(spikes_file, settings)
                figures.append(hst_fig)

            # Average activity
            if "Average activity" in plots:
                if settings.mono_stereo == 0:
                    _, avg_fig = Plots.average_activity(spikes_file, settings)
                else:
                    _, _, avg_fig = Plots.average_activity(spikes_file, settings)
                figures.append(avg_fig)

            # Difference between L/R
            if "Difference between L/R" in plots:
                dlr_fig = Plots.difference_between_LR(spikes_file, settings)
                figures.append(dlr_fig)

            if add_localization_report == True:
                if isinstance(localization_file, LocalizationFile):
                    # MSO spikegram
                    if any("MSO spikegram" in s for s in localization_plots):
                        mso_spikegram = Plots.mso_spikegram(localization_file, settings, localization_settings)
                        pdf.savefig(mso_spikegram)
                        plt.draw()
                    # MSO heatmap
                    if any("MSO heatmap" in s for s in localization_plots):
                        mso_heatmap = Plots.mso_heatmap(localization_file, localization_settings)
                        pdf.savefig(mso_heatmap)
                        plt.draw()
                    # MSO histogram
                    if any("MSO histogram" in s for s in localization_plots):
                        mso_histogram = Plots.mso_histogram(localization_file, settings, localization_settings)
                        pdf.savefig(mso_histogram)
                        plt.draw()
                    # MSO localization
                    if any("MSO localization" in s for s in localization_plots):
                        mso_localization = Plots.mso_localization_plot(localization_file, settings,
                                                                       localization_settings)
                        pdf.savefig(mso_localization)
                        plt.draw()
                else:
                    print("[Functions.PDF_report] > InputFileError: the input LocalizationFile is not valid.")

            for fig in figures:
                pdf.savefig(fig, dpi=150)
                plt.close(fig)

            d = pdf.infodict()
            d['Title'] = 'pyNAVIS report'
            d['Author'] = 'Juan P. Dominguez-Morales'
            d['Subject'] = 'pyNAVIS report'
            d['Keywords'] = 'pyNAVIS'
            d['CreationDate'] = datetime.datetime.today()
            d['ModDate'] = datetime.datetime.today()

            pdf.close()
            print("[Functions.PDF_report] > PDF report generated correctly")

        else:
            print("[Functions.PDF_report] > InputFileError: the input SpikesFile is not valid.")



#################################################################################
##                                                                             ##
##    Copyright C 2018  Juan P. Dominguez-Morales                              ##
##                                                                             ##
##    This file is part of pyNAVIS.                                            ##
##                                                                             ##
##    pyNAVIS is free software: you can redistribute it and/or modify          ##
##    it under the terms of the GNU General Public License as published by     ##
##    the Free Software Foundation, either version 3 of the License, or        ##
##    (at your option) any later version.                                      ##
##                                                                             ##
##    pyNAVIS is distributed in the hope that it will be useful,               ##
##    but WITHOUT ANY WARRANTY; without even the implied warranty of           ##
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the              ##
##    GNU General Public License for more details.                             ##
##                                                                             ##
##    You should have received a copy of the GNU General Public License        ##
##    along with pyNAVIS.  If not, see <http://www.gnu.org/licenses/>.         ##
##                                                                             ##
#################################################################################

import struct
import time
from .utils import Utils

class Savers:

    @staticmethod
    def save_AEDAT(spikes_file, path, settings, verbose = False):
        """
        Saves a SpikesFile into an AEDAT file.

        Parameters:
                spikes_file (SpikesFile): File to save.
                path (string): Path where the output file will be saved, including name. Extension should not be specified.
                settings (MainSettings): Configuration parameters for the file to save.
                verbose (boolean, optional): Set to True if you want the execution time of the function to be printed.

        Returns:
                None.
        """

        if verbose == True: start_time = time.time()
        unpack_param = '>L'
        if settings.address_size == 2:
            unpack_param = ">H"
        elif settings.address_size == 4:
            unpack_param = ">L"

        with open(path + '.aedat', 'wb') as f:
            for i in range(len(spikes_file.addresses)):
                addr = struct.pack(unpack_param, int(spikes_file.addresses[i]))
                ts = struct.pack('>L', int(spikes_file.timestamps[i]//settings.ts_tick))
                f.write(addr)
                f.write(ts)
            if verbose == True:
                print("AEDAT file saved correctly.Took:", time.time() - start_time, 'seconds')


    @staticmethod
    def save_CSV(spikes_file, path, split = False, separator = ', ', verbose = False):
        """
        Saves a SpikesFile into a CSV file where each spike is represented in one line following the same patter: "address, timestamp".

        Parameters:
                spikes_file (SpikesFile): File to save.
                path (string): Path where the output file will be saved, including name. Extension should not be specified.
                split (boolean, optional): True for generating two files (addresses and timestamps), and False for generating one with all the information.
                separator (string, optional): which character to use as separator between addresses and timestamps when split = False
                verbose (boolean, optional): Set to True if you want the execution time of the function to be printed.

        Returns:
                None.
        """

        if verbose == True: start_time = time.time()
        if split == False:
            with open(path + '.csv', 'w') as f:
                for i in range(len(spikes_file.addresses)):
                    f.write(str(spikes_file.addresses[i]) + separator + str(int(spikes_file.timestamps[i])) + "\n")
        else:
            with open(path + '_addrs.csv', 'w') as f:
                for addr in spikes_file.addresses:
                    f.write(str(addr) + "\n")
            with open(path + '_tss.csv', 'w') as f:
                for ts in spikes_file.timestamps:
                    f.write(str(int(ts)) + '\n') 
        if verbose == True:
            print("CSV file saved correctly. Took:", time.time() - start_time, "seconds")


    @staticmethod
    def save_TXT(spikes_file, path, split = True, separator = ', ', verbose = False):
        """
        Saves a SpikesFile into two different TXT files, where addresses and timestamps are stored, respectively.

        Parameters:
                spikes_file (SpikesFile): File to save.
                path (string): Path where the output file will be saved, including name. Extension should not be specified.
                split (boolean, optional): True for generating two files (addresses and timestamps), and False for generating one with all the information.
                separator (string, optional): which character to use as separator between addresses and timestamps when split = False
                verbose (boolean, optional): Set to True if you want the execution time of the function to be printed.

        Returns:
                None.
        """

        if verbose == True: start_time = time.time()
        if split == True:
            with open(path + '_addrs.txt', 'w') as f:
                for addr in spikes_file.addresses:
                    f.write(str(addr) + '\n')
            with open(path + '_tss.txt', 'w') as f:
                for ts in spikes_file.timestamps:
                    f.write(str(int(ts)) + '\n')    
        else:
            with open(path + '.txt', 'w') as f:
                for i in range(len(spikes_file.addresses)):
                    f.write(str(spikes_file.addresses[i]) + separator + str(int(spikes_file.timestamps[i])) + '\n')
        if verbose == True:
            print("TXT file saved correctly. Took:", time.time() - start_time, "seconds")


    @staticmethod
    def save_TXT_relativeTS(spikes_file, path, split = True, separator = ', ', verbose = False):
        """
        Saves a SpikesFile into two different TXT files, where addresses and timestamps are stored, respectively. Timestamps are relative to the previous spike.

        Parameters:
                spikes_file (SpikesFile): File to save.
                path (string): Path where the output file will be saved, including name. Extension should not be specified.
                split (boolean, optional): True for generating two files (addresses and timestamps), and False for generating one with all the information.
                separator (string, optional): which character to use as separator between addresses and timestamps when split = False
                verbose (boolean, optional): Set to True if you want the execution time of the function to be printed.

        Returns:
                None.
        """

        spikes_file_ordered = Utils.order_timestamps(spikes_file)

        if verbose == True: start_time = time.time()

        if split == True:
            with open(path + '_addrs.txt', 'w') as f:
                for addr in spikes_file_ordered.addresses:
                    f.write(str(addr) + '\n')
            with open(path + '_tss.txt', 'w') as f:
                for i in range(len(spikes_file_ordered.timestamps)):
                    if i == 0:
                        f.write(str(0) + '\n')
                    else:
                        f.write(str(int(spikes_file_ordered.timestamps[i]-spikes_file_ordered.timestamps[i-1])) + '\n')
        else:
            with open(path + '_relativeTS.txt', 'w') as f:
                for i in range(len(spikes_file.addresses)):
                    if i == 0:
                        f.write(str(spikes_file.addresses[i]) + separator + str(0) + '\n')
                    else:
                        f.write(spikes_file.addresses[i] + separator + str(int(spikes_file_ordered.timestamps[i]-spikes_file_ordered.timestamps[i-1])) + '\n')
        if verbose == True:
            print("TXT file saved correctly. Took:", time.time() - start_time, "seconds")


    @staticmethod
    def save_as_any(spikes_file, path, output_format, settings=None):
        """
        Saves a SpikesFile into any of the implemented Savers, depending on the output format selected.

        Parameters:
                spikes_file (SpikesFile): File to save.                
                path (string): Path where the output file will be saved. Format should not be specified.
                output_format (string): Output format of the file. Currently supports '.aedat', '.csv', ".txt" and ".txt_rel". See the Savers class for more information.
                settings (MainSettings, optional): Configuration parameters for the output file. Only needed when saving the output as an AEDAT file.

        Returns:
                None.

        Raises:
                SettingsError: if settings are not specified and output_format refers to AEDAT.
        """

        if output_format in ['aedat', 'AEDAT', 'AEDAT', 'AEDAT', 'Aedat', '.aedat']:
            if settings != None:
                Savers.save_AEDAT(spikes_file, path, settings)
            else:
                print('[Savers.save_as_any] > SettingsError: Settings need to be specified when saving the file as an AEDAT filea.')                
        elif output_format in ['csv', 'CSV', '.csv']:
            Savers.save_CSV(spikes_file, path)
        elif output_format in ['txt', 'TXT', '.txt']:
            Savers.save_TXT(spikes_file, path)
        elif output_format in ['txt_rel', 'TXT_rel', '.txt_rel']:
            Savers.save_TXT_relativeTS(spikes_file, path)


#################################################################################
##                                                                             ##
##    Copyright C 2018  Juan P. Dominguez-Morales                              ##
##                                                                             ##
##    This file is part of pyNAVIS.                                            ##
##                                                                             ##
##    pyNAVIS is free software: you can redistribute it and/or modify          ##
##    it under the terms of the GNU General Public License as published by     ##
##    the Free Software Foundation, either version 3 of the License, or        ##
##    (at your option) any later version.                                      ##
##                                                                             ##
##    pyNAVIS is distributed in the hope that it will be useful,               ##
##    but WITHOUT ANY WARRANTY; without even the implied warranty of           ##
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the              ##
##    GNU General Public License for more details.                             ##
##                                                                             ##
##    You should have received a copy of the GNU General Public License        ##
##    along with pyNAVIS.  If not, see <http://www.gnu.org/licenses/>.         ##
##                                                                             ##
#################################################################################

import math
import random
import time
from bisect import bisect_left, bisect_right

import matplotlib.pyplot as plt
import numpy as np
from matplotlib.colors import LinearSegmentedColormap
from .functions import Functions


class Plots:
    @staticmethod
    def spikegram(spikes_file, settings, dot_size=0.2, dot_freq=1, graph_title='Spikegram', start_at_zero=True,
                  verbose=False):
        """
        Plots the spikegram (also known as cochleogram or raster plot) of a SpikesFile.

        This is, a graph where the X axis means time and the Y axis represents addresses (or cochlea channels), and where every spike is plotted as a dot.

        Parameters:
                spikes_file (SpikesFile): File to plot.
                settings (MainSettings): Configuration parameters for the file to plot.
                dot_size (float): Size of the dots used in the spikegram plot.
                dot_freq (int): Set the frequency of spikes that will be represented in the spikegram.
                graph_title (string, optional): Text that will appear as title for the graph.
                start_at_zero (boolean, optional): If set to True, the X axis will start at 0, instead of starting at the minimum timestamp.
                verbose (boolean, optional): Set to True if you want the execution time of the function to be printed.

        Returns:
                None.

        Note:
                A value of 10 in dot_freq means that for every 10 spikes, only 1 will be plotted. This helps reducing lag when plotting heavy files.
        """
        if verbose:
            start_time = time.time()
        # REPRESENTATION
        plt.style.use('seaborn-v0_8-whitegrid')
        spk_fig = plt.figure()
        spk_fig.canvas.manager.set_window_title(graph_title)

        random.seed(0)

        mid_address = settings.num_channels * (settings.on_off_both + 1)

        if start_at_zero and spikes_file.min_ts != 0:
            Functions.adapt_timestamps(spikes_file, settings)

        if settings.mono_stereo == 0:
            plt.scatter(spikes_file.timestamps[0::dot_freq], spikes_file.addresses[0::dot_freq],
                        s=dot_size, rasterized=True)
        else:
            # Convert to numpy arrays
            addresses = np.array(spikes_file.addresses, copy=False)
            timestamps = np.array(spikes_file.timestamps, copy=False)

            sup_indexes = np.argwhere(addresses >= mid_address)

            sup_addresses = addresses[sup_indexes]
            sup_addresses = sup_addresses[::dot_freq]
            sup_timestamps = timestamps[sup_indexes]
            sup_timestamps = sup_timestamps[::dot_freq]

            inf_addresses = np.delete(addresses, sup_indexes)
            inf_addresses = inf_addresses[::dot_freq]
            inf_timestamps = np.delete(timestamps, sup_indexes)
            inf_timestamps = inf_timestamps[::dot_freq]

            plt.scatter(inf_timestamps, inf_addresses, s=dot_size, label="Left cochlea", rasterized=True)
            plt.scatter(sup_timestamps, sup_addresses, s=dot_size, label="Right cochlea", rasterized=True)
            plt.legend(fancybox=False, ncol=2, loc='upper center', markerscale=2 / dot_size, frameon=True)

        if verbose:
            print('SPIKEGRAM CALCULATION', time.time() - start_time)

        plt.title(graph_title, fontsize='x-large')
        plt.xlabel('Timestamp ($\mu$s)', fontsize='large')
        plt.xlim([spikes_file.min_ts, spikes_file.max_ts])
        plt.ylabel('Address', fontsize='large')
        plt.ylim([0, mid_address * (settings.mono_stereo + 1)])

        plt.tight_layout()

        return spk_fig

    @staticmethod
    def sonogram(spikes_file, settings, return_data=False, graph_title='Sonogram', start_at_zero=True, verbose=False):
        """
        Plots the sonogram of a SpikesFile.

        This is, a graph where the X axis means time and the Y axis represents addresses (or cochlea channels), and where the spiking activity is shown with color.

        Parameters:
                spikes_file (SpikesFile): File to plot.
                settings (MainSettings): Configuration parameters for the file to plot.
                return_data (boolean, optional): When set to True, the sonogram matrix will be returned instead of plotted.
                graph_title (string, optional): Text that will appear as title for the graph.
                start_at_zero (boolean, optional): If set to True, the X axis will start at 0, instead of starting at the minimum timestamp.
                verbose (boolean, optional): Set to True if you want the execution time of the function to be printed.

        Returns:
                int[ , ]: Sonogram matrix. Only returned if return_data is set to True.
        """
        if verbose:
            start_time = time.time()

        # In range timestamps
        if start_at_zero and spikes_file.min_ts != 0:
            Functions.adapt_timestamps(spikes_file, settings)

        total_time = spikes_file.max_ts - spikes_file.min_ts
        last_time = spikes_file.min_ts

        # Calculate number of addresses and number of windows
        num_addresses = settings.num_channels * (settings.on_off_both + 1) * (settings.mono_stereo + 1)
        num_windows = int(math.ceil(total_time / settings.bin_size))

        # Define sonogram array
        sonogram = np.zeros((num_addresses, num_windows))

        # Bincount
        for i in range(num_windows):
            a = bisect_left(spikes_file.timestamps, last_time)
            b = bisect_right(spikes_file.timestamps, last_time + settings.bin_size)

            blockAddr = spikes_file.addresses[a:b]

            spikes = np.bincount(blockAddr, minlength=num_addresses)

            last_time += settings.bin_size
            sonogram[:, i] = spikes

        if verbose:
            print('SONOGRAM CALCULATION', time.time() - start_time)

        if not return_data:
            # REPRESENTATION
            plt.style.use('default')
            sng_fig = plt.figure()
            sng_fig.canvas.manager.set_window_title(graph_title)

            plt.imshow(sonogram, aspect="auto", cmap='hot', rasterized=True)  # , aspect="auto")
            plt.gca().invert_yaxis()

            plt.xlabel('Bin (' + str(settings.bin_size) + '$\mu$s width)', fontsize='large')
            plt.ylabel('Address', fontsize='large')

            plt.title(graph_title, fontsize='x-large')

            """
            plt.annotate('Right cochlea | Left cochlea',
            xy=(1.00, 0.5), xytext=(5, 0),
            xycoords=('axes fraction', 'figure fraction'),
            textcoords='offset points',
            size=11, ha='center', va='center', rotation=270)
            """

            if settings.mono_stereo == 1:
                plt.annotate('Right cochlea',
                             xy=(1.00, 0.75), xytext=(5, 0),
                             xycoords=('axes fraction', 'axes fraction'),
                             textcoords='offset points',
                             size=11, ha='center', va='center', rotation=270)

                plt.annotate('Left cochlea',
                             xy=(1.00, 0.25), xytext=(5, 0),
                             xycoords=('axes fraction', 'axes fraction'),
                             textcoords='offset points',
                             size=11, ha='center', va='center', rotation=270)

            colorbar = plt.colorbar()
            colorbar.set_label('No. of spikes', rotation=270, fontsize='large', labelpad=10)

            return sng_fig
        else:
            return sonogram

    @staticmethod
    def histogram(spikes_file, settings, bar_line=1, graph_title='Histogram', verbose=False):
        """
        Plots the histogram of a SpikesFile.

        This is, a graph where addresses (or cochlea channels) are represented in the X axis, and number of spikes in the Y axis.

        Parameters:
                spikes_file (SpikesFile): File to plot.
                settings (MainSettings): Configuration parameters for the file to plot.
                bar_line (int, optional): Select wether to plot the histogram as bar plot (0) or as a line graph (1).
                graph_title (string, optional): Text that will appear as title for the graph.
                verbose (boolean, optional): Set to True if you want the execution time of the function to be printed.

        Returns:
                int[]: Histogram array.
        """

        start_time = time.time()

        spikes_count = np.bincount(spikes_file.addresses,
                                   minlength=settings.num_channels * (settings.on_off_both + 1) * (settings.mono_stereo + 1))

        if verbose:
            print('HISTOGRAM CALCULATION:', time.time() - start_time)

        plt.style.use('seaborn-v0_8-whitegrid')
        hst_fig = plt.figure()
        hst_fig.canvas.manager.set_window_title(graph_title)
        plt.title(graph_title, fontsize='x-large')
        plt.xlabel('Address', fontsize='large')
        plt.ylabel('No. of spikes', fontsize='large')

        if bar_line == 0:
            if settings.mono_stereo == 1:
                plt.bar(np.arange(settings.num_channels * (settings.on_off_both + 1)),
                        spikes_count[0:settings.num_channels * (settings.on_off_both + 1)],
                        rasterized=True)
                plt.bar(np.arange(settings.num_channels * (settings.on_off_both + 1)),
                        spikes_count[settings.num_channels * (settings.on_off_both + 1):settings.num_channels * 2 * (settings.on_off_both + 1)],
                        rasterized=True)
            else:
                plt.bar(np.arange(settings.num_channels * (settings.on_off_both + 1) * (settings.mono_stereo + 1)),
                        spikes_count, rasterized=True)
        else:
            if settings.mono_stereo == 1:
                plt.plot(np.arange(settings.num_channels * (settings.on_off_both + 1)),
                         spikes_count[0:settings.num_channels * (settings.on_off_both + 1)],
                         label='Left cochlea', rasterized=True)
                plt.plot(np.arange(settings.num_channels * (settings.on_off_both + 1)),
                         spikes_count[settings.num_channels * (settings.on_off_both + 1):settings.num_channels * 2 * (settings.on_off_both + 1)],
                         label='Right cochlea', rasterized=True)
                plt.legend(loc='best', frameon=True)
            else:
                plt.plot(np.arange(settings.num_channels * (settings.on_off_both + 1) * (settings.mono_stereo + 1)),
                         spikes_count, rasterized=True)

        plt.tight_layout()

        return spikes_count, hst_fig

    @staticmethod
    def average_activity(spikes_file, settings, graph_title='Average activity', verbose=False):
        """
        Plots the average activity plot of a SpikesFile.

        This is, a graph where time is represented in the X axis, and average number of spikes in the Y axis.

        Parameters:
                spikes_file (SpikesFile): File to plot.
                settings (MainSettings): Configuration parameters for the file to plot.
                graph_title (string, optional): Text that will appear as title for the graph.
                verbose (boolean, optional): Set to True if you want the execution time of the function to be printed.

        Returns:
                int[ ] average_activity_L: Average activity array.
                int[ ] average_activity_R: Average activity array. Only returned if the mono_stereo parameter in settings is set to 1
        """
        # Convert to numpy array
        addresses = np.array(spikes_file.addresses, copy=False)
        timestamps = np.array(spikes_file.timestamps, copy=False)

        # Define plot variables
        total_time = spikes_file.max_ts - spikes_file.min_ts
        mid_address = settings.num_channels * (settings.on_off_both + 1)
        num_bins = int(math.ceil(total_time / settings.bin_size))

        average_activity_L = np.zeros(num_bins)
        if settings.mono_stereo == 1:
            average_activity_R = np.zeros(num_bins)

        if verbose:
            start_time = time.time()

        # Array of bins
        bins = np.arange(num_bins) * settings.bin_size

        # Calculate the bin associated with each timestamp
        bins_indexes = np.digitize(timestamps, bins)

        # Cutting indexes (timestamps must be sorted)
        cut_indexes = np.where(np.diff(bins_indexes) > 0)[0]

        # Split timestamps by cutting indexes
        spikes_per_bins = np.array_split(addresses, cut_indexes)

        # Calculate the number of spikes in the bins
        for i in range(len(spikes_per_bins)):
            count_below = np.count_nonzero(spikes_per_bins[i] < mid_address)
            average_activity_L[i] = count_below
            if settings.mono_stereo == 1:
                average_activity_R[i] = len(spikes_per_bins[i]) - count_below

        if verbose:
            print('AVERAGE ACTIVITY CALCULATION', time.time() - start_time)

        plt.style.use('seaborn-v0_8-whitegrid')
        avg_fig = plt.figure()
        avg_fig.canvas.manager.set_window_title(graph_title)
        plt.title(graph_title, fontsize='x-large')
        plt.xlabel('Bin (' + str(settings.bin_size) + '$\mu$s width)', fontsize='large')
        plt.ylabel('No. of spikes', fontsize='large')

        plt.plot(bins / settings.bin_size, average_activity_L, label='Left cochlea', rasterized=True)
        if settings.mono_stereo == 1:
            plt.plot(bins / settings.bin_size, average_activity_R, label='Right cochlea')
            plt.legend(loc='best', ncol=2, frameon=True)

        plt.tight_layout()
        if settings.mono_stereo == 0:
            return average_activity_L, avg_fig
        else:
            return average_activity_L, average_activity_R, avg_fig

    @staticmethod
    def difference_between_LR(spikes_file, settings, return_data=False, graph_title='Diff. between L and R cochlea',
                              verbose=False, start_at_zero=True):
        """
        Plots a plot showing the differente between the left and the right activity of a SpikesFile.

        Parameters:
                spikes_file (SpikesFile): File to plot.
                settings (MainSettings): Configuration parameters for the file to plot.
                return_data (boolean, optional): When set to True, the sonogram matrix will be returned instead of plotted.
                graph_title (string, optional): Text that will appear as title for the graph.
                verbose (boolean, optional): Set to True if you want the execution time of the function to be printed.
                start_at_zero (boolean, optional): If set to True, the X axis will start at 0, instead of starting at the minimum timestamp.

        Returns:
                int[ , ]: Disparity matrix. Only returned if return_data is set to True.

        Raises:
                SettingsError: if settings.mono_stereo == 0

        Note:
                This function can only be called if the mono_stereo parameter in settings is set to 1.
        """
        if settings.mono_stereo == 1:
            if verbose:
                start_time = time.time()

            # In range timestamps
            if start_at_zero and spikes_file.min_ts != 0:
                Functions.adapt_timestamps(spikes_file, settings)

            total_time = spikes_file.max_ts - spikes_file.min_ts
            last_time = spikes_file.min_ts

            # Calculate number of addresses and number of windows
            num_addresses = settings.num_channels * (settings.on_off_both + 1) * (settings.mono_stereo + 1)
            mid_address = round(num_addresses / 2)
            num_windows = int(math.ceil(total_time / settings.bin_size))

            # Define diff array and partial arrays
            diff = np.zeros((round(num_addresses / 2), num_windows))

            if verbose:
                start_time = time.time()

            for i in range(num_windows):
                a = bisect_left(spikes_file.timestamps, last_time)
                b = bisect_right(spikes_file.timestamps, last_time + settings.bin_size)

                blockAddr = spikes_file.addresses[a:b]

                spikes = np.bincount(blockAddr, minlength=num_addresses)

                diff[:, i] = spikes[0:mid_address] - spikes[mid_address:num_addresses]

                last_time += settings.bin_size

            if verbose:
                print('DIFF CALCULATION', time.time() - start_time)

            max_abs = np.max(abs(diff))
            if max_abs != 0:
                diff = diff * 100 / max_abs

            if not return_data:
                # REPRESENTATION
                plt.style.use('default')
                dlr_fig = plt.figure()
                dlr_fig.canvas.manager.set_window_title(graph_title)

                # cmap = 'RdBu'
                colors = [(1, 0.49803921568627450980392156862745, 0.05490196078431372549019607843137), (1, 1, 1), (
                    0.12156862745098039215686274509804, 0.46666666666666666666666666666667,
                    0.70588235294117647058823529411765)]  # R -> G -> B
                n_bins = [3, 6, 10, 100]  # Discretizes the interpolation into bins
                cmap_name = 'my_list'
                cm = LinearSegmentedColormap.from_list(cmap_name, colors, N=100)

                plt.imshow(diff, vmin=-100, vmax=100, aspect="auto", cmap=cm, rasterized=True)  # , aspect="auto")
                plt.gca().invert_yaxis()

                plt.xlabel('Bin (' + str(settings.bin_size) + '$\mu$s width)', fontsize='large')
                plt.ylabel('Address', fontsize='large')

                plt.title(graph_title, fontsize='x-large')

                colorbar = plt.colorbar(ticks=[100, 50, 0, -50, -100], orientation='horizontal')
                colorbar.set_label('Cochlea predominance', rotation=0, fontsize='large', labelpad=10)
                colorbar.ax.set_xticklabels(['100% L\nCochlea', '50%', '0%\nL==R', '50%', '100% R\nCochlea'])
                colorbar.ax.invert_xaxis()

                return dlr_fig
            else:
                return diff
        else:
            # print("This functionality is only available for stereo AEDAT files.")
            print("[Plots.difference_between_LR] > SettingsError: This functionality is only available for stereo files.")

    @staticmethod
    def mso_heatmap(localization_file, localization_settings, graph_title="MSO heatmap", enable_colorbar=True,
                    verbose=False):
        """
        Plots the heatmap for the MSO activity extracted from a LocalizationFile.

        This is, a graph where the X axis means the neuron ID, the Y axis means the frequency channel to which the MSO neuron's population are connected, and the color means the activity.

        Parameters:
                localization_file (LocalizationFile): Localization file to plot.
                localization_settings (LocalizationSettings): Localization configuration parameters for the file to plot.
                graph_title (string, optional): Text that will appear as title for the graph.
                enable_colorbar (boolean, optional): Set to True if you want to show the color bar that indicates the activity range by colors.
                verbose (boolean, optional): Set to True if you want the execution time of the function to be printed.

        Returns:
                None.

        Note:
                None.
        """

        if verbose == True: start_time = time.time()

        # Get the number of frequency channels set in the configuration
        mso_number_freq_ch = localization_settings.mso_end_channel - localization_settings.mso_start_channel + 1

        # Create the activity matrix
        mso_activity = np.zeros((mso_number_freq_ch, localization_settings.mso_num_neurons_channel))

        num_mso_events = len(localization_file.mso_timestamps)

        for i in range(0, num_mso_events):
            # Move the frequency channel from the relative range to an absolute range starting at zero
            freq_channel = localization_file.mso_channels[i] - localization_settings.mso_start_channel
            neuron_id = localization_file.mso_neuron_ids[i]
            # Accumulate the activity for each neuron for each frequency channel according to the LocalizationFIle
            mso_activity[freq_channel][neuron_id] = mso_activity[freq_channel][neuron_id] + 1

        if verbose == True: print('MSO HEATMAP CALCULATION', time.time() - start_time)

        # Generate the labels lists
        freq_channel_labels = []
        for i in range(localization_settings.mso_start_channel, localization_settings.mso_end_channel + 1):
            freq_channel_labels.append(str(i))

        neuron_id_labels = []
        for i in range(0, localization_settings.mso_num_neurons_channel):
            neuron_id_labels.append(str(i))

        # REPRESENTATION
        plt.style.use('seaborn-ticks')
        htmap_fig, htmap_ax = plt.subplots()
        htmap_fig.canvas.manager.set_window_title(graph_title)

        # Create the heatmap image
        htmap_im = plt.imshow(mso_activity, cmap='viridis')

        if enable_colorbar == True:
            colorbar = plt.colorbar()
            colorbar.set_label('No. of spikes', rotation=270, fontsize='large', labelpad=10)

        # Set the ticks
        htmap_ax.set_xticks(np.arange(len(neuron_id_labels)))
        htmap_ax.set_yticks(np.arange(len(freq_channel_labels)))
        # And set all the ticks' labels
        htmap_ax.set_xticklabels(neuron_id_labels)
        htmap_ax.set_yticklabels(freq_channel_labels)

        htmap_ax.set_xlabel('Neuron ID', fontsize='large')
        htmap_ax.set_ylabel('Freq. channel', fontsize='large')

        # Rotate the tick labels and set their alignment.
        plt.setp(htmap_ax.get_xticklabels(), rotation=45, ha="right",
                 rotation_mode="anchor")

        # Loop over data dimensions and create text annotations.
        for i in range(len(freq_channel_labels)):
            for j in range(len(neuron_id_labels)):
                text = htmap_ax.text(j, i, mso_activity[i, j],
                                     ha="center", va="center", color="w", fontsize='xx-small')

        plt.title(graph_title, fontsize='x-large')

        plt.tight_layout()
        htmap_fig.show()

    @staticmethod
    def mso_spikegram(localization_file, settings, localization_settings, dot_size=0.2, graph_title='MSO spikegram',
                      verbose=False):
        """
        Plots the 3D spikegram (also known as raster plot) of the MSO information contained in a LocalizationFile.
        This is, a graph where the X axis represents neuron IDs, the Y axis means time, and the Z axis represents the frequency channels of the cochlea, and where every spike is plotted as a dot.

        Parameters:
                localization_file (LocalizationFile): Localization file to plot.
                settings (MainSettings): Configuration parameters for the file to plot.
                localization_settings (LocalizationSettings): Localization configuration parameters for the file to plot.
                dot_size (float): Size of the dots used in the spikegram plot.
                graph_title (string, optional): Text that will appear as title for the graph.
                verbose (boolean, optional): Set to True if you want the execution time of the function to be printed.

        Returns:
                None.

        Note:
                None.
        """

        if verbose == True: start_time = time.time()

        # REPRESENTATION
        plt.style.use('seaborn-v0_8-whitegrid')
        msospk_fig = plt.figure()
        ax = msospk_fig.add_subplot(111, projection='3d')
        msospk_fig.canvas.manager.set_window_title(graph_title)

        # Plot all the spikes stored in the localization_file
        ax.scatter(localization_file.mso_neuron_ids, localization_file.mso_timestamps, localization_file.mso_channels,
                   s=dot_size)

        if verbose == True: print('MSO SPIKEGRAM CALCULATION', time.time() - start_time)

        plt.title(graph_title, fontsize='x-large')

        # Set the label of each axis
        ax.set_xlabel('Neuron ID', fontsize='large')
        ax.set_ylabel('Timestamp ($\mu$s)', fontsize='large')
        ax.set_zlabel('Freq. channel', fontsize='large')

        # Set the axis' limits for each axis according to the LocalizationSettings parameters
        ax.set_xlim([0, localization_settings.mso_num_neurons_channel])
        ax.set_ylim([0, localization_file.mso_timestamps[-1]])
        ax.set_zlim([0, settings.num_channels])

        plt.tight_layout()
        msospk_fig.show()

    @staticmethod
    def mso_localization_plot(localization_file, settings, localization_settings,
                              graph_title="MSO localization estimation", start_at_zero=True, verbose=False):
        """
        Plots the result of the coincidence counters of the Jeffress model for the MSO according to the activity in the LocalizationFile.

        This is, the neuron that fired the most in a time bin, thus indicating the sound source position.

        Parameters:
                localization_file (LocalizationFile): Localization file to plot.
                settings (MainSettings): Configuration parameters for the file to plot.
                localization_settings (LocalizationSettings): Localization configuration parameters for the file to plot.
                graph_title (string, optional): Text that will appear as title for the graph.
                start_at_zero (boolean, optional): If set to True, the X axis will start at 0, instead of starting at the minimum timestamp.
                verbose (boolean, optional): Set to True if you want the execution time of the function to be printed.

        Returns:
                None.

        Note:
                None.
        """

        if verbose == True: start_time = time.time()

        last_time = 0

        # Set the timestamps limits according to start_at_zero option
        if start_at_zero:
            total_time = max(localization_file.mso_timestamps)
            last_time = 0
        else:
            total_time = max(localization_file.mso_timestamps) - min(localization_file.mso_timestamps)
            last_time = min(localization_file.mso_timestamps)

        # Estimate the number of time bins
        its = int(math.ceil(total_time / settings.bin_size))

        # Create the activity array
        mso_activity = np.zeros((its, localization_settings.mso_num_neurons_channel))
        mso_max_activity = np.zeros(its)

        # For each time bin, calculate the activity of all the MSO neurons and get the winner
        for i in range(0, its):
            a = bisect_left(localization_file.mso_timestamps, last_time)
            b = bisect_right(localization_file.mso_timestamps, last_time + settings.bin_size)

            blockNeuronIDs = localization_file.mso_neuron_ids[a:b]

            spikes = np.bincount(blockNeuronIDs, minlength=localization_settings.mso_num_neurons_channel)

            last_time += settings.bin_size

            mso_activity[i, :] = spikes

            index_max_activity = np.argmax(mso_activity[i])

            mso_max_activity[i] = index_max_activity

        if verbose == True: print('MSO_LOCALIZATION PLOT CALCULATION', time.time() - start_time)

        # Set the figure
        plt.style.use('seaborn-v0_8-whitegrid')
        mso_loc_fig = plt.figure()
        mso_loc_fig.canvas.manager.set_window_title(graph_title)
        plt.title(graph_title, fontsize='x-large')
        plt.xlabel('Bin (' + str(settings.bin_size) + '$\mu$s width)', fontsize='large')
        plt.ylabel('Position (in degrees)', fontsize='large')
        plt.ylim([-1, localization_settings.mso_num_neurons_channel + 1])

        yticklabels = []
        angle_slot = 180.0 / localization_settings.mso_num_neurons_channel
        for slot_index in range(0, localization_settings.mso_num_neurons_channel):
            middle_angle = (slot_index * angle_slot) - 90.0 + (angle_slot / 2.0)
            yticklabels.append(str(int(abs(middle_angle))))

        plt.yticks(ticks=np.arange(localization_settings.mso_num_neurons_channel), labels=yticklabels, fontsize='small')

        plt.plot(np.arange(math.ceil(total_time / settings.bin_size)), mso_max_activity, label='Position estimation')

        # Add some text to clarify the results
        plt.annotate('Left side',
                     xy=(1.00, 0.85), xytext=(5, 0),
                     xycoords=('axes fraction', 'axes fraction'),
                     textcoords='offset points',
                     size=11, ha='center', va='center', rotation=270)

        plt.annotate('Centre',
                     xy=(1.00, 0.50), xytext=(5, 0),
                     xycoords=('axes fraction', 'axes fraction'),
                     textcoords='offset points',
                     size=11, ha='center', va='center', rotation=270)

        plt.annotate('Right side',
                     xy=(1.00, 0.15), xytext=(5, 0),
                     xycoords=('axes fraction', 'axes fraction'),
                     textcoords='offset points',
                     size=11, ha='center', va='center', rotation=270)

        plt.tight_layout()
        mso_loc_fig.show()

    @staticmethod
    def mso_histogram(localization_file, settings, localization_settings, graph_title='MSO histogram', verbose=False):
        """
        Plots the 3D histogram of the MSO information contained in a LocalizationFile.

        This is, a graph where neuron IDs are represented in the X axis, frequency channels are represented in the Y axis, and number of spikes in the Z axis.

        Parameters:
                localization_file (LocalizationFile): Localization file to plot.
                settings (MainSettings): Configuration parameters for the file to plot.
                localization_settings (LocalizationSettings): Localization configuration parameters for the file to plot.
                graph_title (string, optional): Text that will appear as title for the graph.
                verbose (boolean, optional): Set to True if you want the execution time of the function to be printed.

        Returns:
                None.

        Note:
                None.
        """

        start_time = time.time()

        # Set the total number of frequency channels
        mso_number_freq_ch = localization_settings.mso_end_channel - localization_settings.mso_start_channel + 1

        # Create the activity matrix
        mso_activity = np.zeros((mso_number_freq_ch, localization_settings.mso_num_neurons_channel))

        num_mso_events = len(localization_file.mso_timestamps)

        # Calculate the total number of events for each frequency channel and for each neuron ID
        for i in range(0, num_mso_events):
            freq_channel = localization_file.mso_channels[i] - localization_settings.mso_start_channel
            neuron_id = localization_file.mso_neuron_ids[i]
            # Accumulate the activity for each neuron for each frequency channel according to the LocalizationFIle
            mso_activity[freq_channel][neuron_id] = mso_activity[freq_channel][neuron_id] + 1

        if verbose == True: print('MSO HISTOGRAM CALCULATION:', time.time() - start_time)

        # Representation
        plt.style.use('seaborn-v0_8-whitegrid')
        mso_hst_fig = plt.figure()
        ax = mso_hst_fig.add_subplot(111, projection='3d')

        # Set the data to show
        _x = np.arange(localization_settings.mso_num_neurons_channel)
        _y = np.arange(mso_number_freq_ch)
        _xx, _yy = np.meshgrid(_x, _y)
        x, y = _xx.ravel(), _yy.ravel()
        z = mso_activity.ravel()

        top = z
        bottom = np.zeros_like(top)
        width = depth = 1

        ax.bar3d(x, y, bottom, width - 0.25, depth - 0.5, top, shade=True)
        ax.set_title('Shaded')

        # Set the axes' limits
        ax.set_xlim3d(0, localization_settings.mso_num_neurons_channel)
        ax.set_ylim3d(0, mso_number_freq_ch)
        ax.set_zlim3d(0, np.amax(z))

        # Set the axes' labels
        ax.set_xlabel('Neuron ID', fontsize='large')
        ax.set_ylabel('Freq. channel', fontsize='large')
        ax.set_zlabel('No. of spikes', fontsize='large')

        # Generate the labels lists
        freq_channel_labels = []
        for i in range(localization_settings.mso_start_channel, localization_settings.mso_end_channel + 1):
            freq_channel_labels.append(str(i))

        # Set all the ticks
        ax.set_yticks(np.arange(len(freq_channel_labels)))
        # And all the ticks' labels
        ax.set_yticklabels(freq_channel_labels)

        mso_hst_fig.canvas.manager.set_window_title(graph_title)
        plt.title(graph_title, fontsize='x-large')

        mso_hst_fig.show()



from .functions import Functions
from .report_functions import ReportFunctions
from .generators import Generators
from .loaders import Loaders
from .objects import SpikesFile
from .plots import Plots
from .savers import Savers
from .splitters import Splitters
from .utils import Utils
from .main_settings import MainSettings, LocalizationSettings
from .dataset_gen import DatasetGenerators


#################################################################################
##                                                                             ##
##    Copyright C 2018  Juan P. Dominguez-Morales                              ##
##                                                                             ##
##    This file is part of pyNAVIS.                                            ##
##                                                                             ##
##    pyNAVIS is free software: you can redistribute it and/or modify          ##
##    it under the terms of the GNU General Public License as published by     ##
##    the Free Software Foundation, either version 3 of the License, or        ##
##    (at your option) any later version.                                      ##
##                                                                             ##
##    pyNAVIS is distributed in the hope that it will be useful,               ##
##    but WITHOUT ANY WARRANTY; without even the implied warranty of           ##
##    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.See the              ##
##    GNU General Public License for more details.                             ##
##                                                                             ##
##    You should have received a copy of the GNU General Public License        ##
##    along with pyNAVIS.  If not, see <http://www.gnu.org/licenses/>.         ##
##                                                                             ##
#################################################################################

class MainSettings:
    """
    Class that collects the main configuration settings of pyNAVIS

    Attributes:
        mono_stereo (int): Set to 0 for mono files and 1 for stereo files.
        bin_size (int): Bin width (or window size) to use when processing the information.
        ts_tick (float): Timestamp tick. Correspondence factor between timestamp value in file and actual time.
        num_channels (int): Number of cochlea channels.
        address_size (int): Number of bytes that each address is using. Only needed when loading .aedat files
        timestamp_size (int): Number of bytes that each timestamp is using. Only needed when loading .aedat files
        on_off_both (int): Select wether the addresses contained in the file are ON, OFF or if it is using both.
        reset_timestamp (boolean): Select wether to have timedtamps starting at 0 (True) or leave it as they are (False).
        verbose (boolean): Select wether to print information about the file being loaded (True) or not (False
    
    Notes:
            Set ts_tick to 1 for .aedat files recorded with jAER, to 0.2 for files recorded with USBAERmini2, and to 80e-3 for files recorded with zynqGrabber.

            Set address_size to 2 for .eadat files recorded with USBAERmini2, or to 4 for files recorded with jAER.
            This parameter is only relevant for loading AEDAT (.aedat) files.

            Set on_off_both to 0 if addresses are only ON or OFF, or to 1 if using both ON and OFF.

            reset_timestamp subtracts the smallest timestamp in the file to each of the timestamps.
    """

    def __init__(self, num_channels, mono_stereo = 0, address_size = 2, timestamp_size=4, ts_tick = 1, bin_size = 20000, on_off_both = 1, reset_timestamp = True,verbose=True):
        self.num_channels = num_channels
        self.mono_stereo = mono_stereo
        self.address_size = address_size
        self.timestamp_size = timestamp_size
        self.ts_tick = ts_tick
        self.bin_size = bin_size
        self.on_off_both = on_off_both
        self.reset_timestamp = reset_timestamp
        self.verbose = verbose
        
class LocalizationSettings:
    """
    Class that collects the configuration settings of pyNAVS when working with a NAS model that integrates the sound source localization model.

    Attributes:
        mso_start_channel (int): First NAS frequency channel that is connected to the MSO module.
        mso_end_channel (int): Last NAS frequency channel that is connected to the MSO module.
        mso_num_neurons_channel (int): Number of coincidence detector neurons used for each frequency channel in the MSO module.
        lso_start_channel (int): First NAS frequency channel that is connected to the LSO module.
        lso_end_channel (int): Last NAS frequency channel that is connected to the LSO module.
        lso_num_neurons_channel (int): Number of neurons used for each frequency channel in the LSO module.

    Notes:
            The first frequency channel is channel number 0.

            The frequency channel range should match the parameter num_channels set in MainSettings.

            The values set here should be the same that the one used for generating the FPGA configuration file.
    """

    def __init__(self, mso_start_channel, mso_end_channel, mso_num_neurons_channel, lso_start_channel = 0, lso_end_channel = 1, lso_num_neurons_channel = 1):
        self.mso_start_channel = mso_start_channel
        self.mso_end_channel = mso_end_channel
        self.mso_num_neurons_channel = mso_num_neurons_channel
        self.lso_start_channel = lso_start_channel
        self.lso_end_channel = lso_end_channel
        self.lso_num_neurons_channel = lso_num_neurons_channel